/* ReSolve V0.08.09g 2023/02/08 solve math expressions using discrete values*/
/* Copyright 2005-2023 Valerio Messina http://users.iol.it/efa              */
/* reSolveCli.c is part of ReSolve
   ReSolve is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   ReSolve is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with ReSolve. If not, see <http://www.gnu.org/licenses/>.          */

/* reSolveCli.c This is the main CLI source */

#include "reSolveLib.h"    /* main source include */

int main (int numPar, char* param[]) {
   u32  first;
   u08  len, pos;
   u08  p;
   char c;
   u16  valTy, resTy;
   float allocatedMB;
   int  ret;

   // Use of dynamic memory allocation
   // 0 - load configuration file and params
   ret = fillConfigVars (); // load and check users config file
   if (ret!=OK) {
      if (dbgLev>=PRINTERROR) printf ("ERROR %s: fillConfigVars returned not OK\n", __FUNCTION__);
      return (ERROR);
   }

   // 1 - read and set user request. As now are still partially hard coded
   if (Eseries>0) {
      numR=(u32)Eseries*decades; /* number of existant values of resistance */
      listNumber=0; // not used with Eseries
   } else { /* custom list */
      // listNumber=57; /* insert here custom list quantity */
      // listNumber=473; /* insert here custom list quantity */
      // listNumber=33; /* insert here custom list quantity */
      // listNumber=14; /* insert here custom list quantity */
      // listNumber=20; /* insert here custom list quantity */
      // listNumber=47; /* insert here custom list quantity */
      // listNumber=14; /* insert here custom list quantity */
      // listNumber=120; /* insert here custom list quantity */
      numR=(u32)listNumber; /* number of existant values of resistances */
   }
   //maxRp=1; /* max number of resistances supported per position: as now 1 or 2 */

   // 2 - calculate the needed memory
   if (maxRp==1) {
      numV=numR; /* possible values x each position */
   }
   if (maxRp==2) {
      numV=numR+numR*numR+numR; /* number of possible values x each position */
   }
   //maxRc=2; /* number of resistances (variables) in the circuit: 2 */
   if (maxRc==2) {   /* as now 2 is the only supported value */
      totV=((u32)numV)*((u32)numV); /* number of values to try */
      // compile&doNotStart with 3'882'988'894 max. Start with 917'797'375 max.
      // compile&doNotStart with   268'402'689 max. Start with 59'954'049 max.
      // RunWell                                          with 59'954'049 max.
   }
   valTy = sizeof(struct rValuesTy);
   u32 rValueSize=numV*valTy; // expected no more than '1344x48=64512'
   if (dbgLev>=PRINTDEBUG) printf ("numV:%u X valTy:%u = rValueSize:%u\n", numV, valTy, rValueSize);

   resTy = sizeof(struct resultsTy);
   u64 resultSize=(u64)totV*resTy;
   if (dbgLev>=PRINTDEBUG) printf ("totV:%u X resTy:%u = resultSize:%llu\n", totV, resTy, resultSize);
   //printf ("Will allocate about %f MB of RAM for val\n", rValueSize/1048576.0);
   //printf ("size of struct resultsTy:%u\n", resTy);
   //printf ("Will allocate about %f MB of RAM for res\n", resultSize/1048576.0);
   allocatedMB = (rValueSize/1048576.0+resultSize/1048576.0);

   // 2b - checking arguments syntax
   /*printf ("numPar:%u\n", numPar);*/
   /*printf ("float:%u, double:%u\n", sizeof(float), sizeof (double));*/
   numPar--;
   /*printf ("numPar:%u\n", numPar);*/
   if (numPar>0) {
      if (!strcmp(param[1],"-h") || !strcmp(param[1],"--help")) {
         showHelp (allocatedMB);
         return (ERROR);
      }
   }
   if (param[1] != NULL) {
      /*printf ("param1Str:%s\n", param[1]);*/
      desired = atof (param[1]);
      /*printf ("param1Num:%f\n", desired);*/
   }
   if (desired==0.0) desired = DesiredDefault; /* letters case */
   if (numPar>1) { /* at least searched value and formula are present */
      if (param[2] != NULL) {
         /*printf ("param2Str:%s\n", param[2]);*/
         len = strlen (param[2]);
         if (len < 60) {
            pos = strcspn (param[2], "!\"#$&',:;<=>?@[\\]_`{|}~簧");
            if (pos!=len) { // filter out: !"#$&',:;<=>?@[\]_`{|}~簧
               showHelp (allocatedMB);
               printf ("Unsupported char: '%c'\n", *(param[2]+pos));
               return (ERROR);
            }
            for (p=0; p<len; p++) {
               c = *(param[2]+p);
               if ((u08)c>127) {
                  showHelp (allocatedMB);
                  printf ("Unsupported char: '%c'\n", *(param[2]+pos));
                  return (ERROR);
               }
               if (isalpha ((u08)c)) { /* with gcc2.95 is required an integer */
                  if ( (toupper ((u08)c)-'A') >= MaxRc ) {
                              /* ^^^^ old gcc2.95 need int ^^^^ */
                     showHelp (allocatedMB);
                     printf ("ERROR: Unsupported Variable:'%c'\n\n", c);
                     return (ERROR);
                  }
               }
            }
            expr = param[2];
         } else {
            showHelp (allocatedMB);
            printf ("Formula too long: %u chars\n", len);
            return (ERROR);
         }
      }
      if (numBestRes==0) numBestRes = NumberResDefault; /* letters case */
   }
   if (numPar>2) { /* searched value, formula and number of results are present */
      if (param[3] != NULL) {
         /*printf ("param3Str:%s\n", param[3]);*/
         numBestRes = atoi (param[3]);
         /*printf ("param3Num:%u\n", numBestRes);*/
      }
      if (numBestRes==0) numBestRes = NumberResDefault; /* letters case */
      if (numPar>3) printf ("WARNING: Ignoring numPar[%u++]='%s'\n", numPar, param[4]);
   }
   showHead ();
   printf ("Found and loaded config file: 'reSolveConf.txt'\n");

   // 3 - allocate the memory asking to the OS a malloc()
   /* allocated memory for all R: [12*7]+([12*7]*[12*7])+[12*7] : */
   //struct rValuesTy rValues[numV]; /* single, series & parallel */
   //struct rValuesTy* rValues; /* pointer to memory for single, series & parallel */
   printf ("Allocating about %f MB of RAM for inputs ...\n", rValueSize/1048576.0);
   rValues=malloc(rValueSize); // numV*sizeof(struct rValuesTy)
   // allocation fail with 2'930'000'000 on a 4 GB RAM, 8 GB swap, 32 bit OS
   if (!rValues) {
      printf ("Dynamic allocation of:%u Bytes failed:0x%8p\n", rValueSize, rValues);
      return (ERROR);
   }
   // 4 - create the structure's vector inside the allocated memory
   for (u32 e=0; e<numV; e++) {
      rValues[e].rp = (double*) malloc (maxRp*sizeof (double));
      if (!rValues[e].rp) {
         printf ("Dynamic allocation of:%zu Bytes failed:0x%8p\n", maxRp*sizeof (double), rValues[e].rp);
         return (ERROR);
      }
   }
   // now can use global rValues[].rp[], rValues[].r for input values
   printf ("Allocating about %f MB of RAM for results ...\n", resultSize/1048576.0);
   //printf ("size_t size:%u Bytes ptr:%lu\n", sizeof (size_t), (u32) pow(2, 8*sizeof (size_t)));
   if (resultSize > pow(2, 8*sizeof (size_t))) {
      printf ("ERROR: This machine cannot handle allocation of:%llu Bytes\n", resultSize);
      return (ERROR);
   }
   results=malloc(resultSize); // (u64)totV*sizeof(struct resultsTy)
   if (dbgLev>=PRINTDEBUG) printf ("allocated %llu Bytes at:0x%8p\n", resultSize, results);
   if (!results) {
      printf ("Dynamic allocation of:%llu Bytes failed:0x%8p\n", resultSize, results);
      return (ERROR);
   }
   // now can use global results[].pos[], results[].delta for results
   if (dbgLev>=PRINTDEBUG) printf ("Done. Dynamically allocated 'results' at:%p\n", results);
   printf ("Desired Value  :%11.5f\n", desired);
   printf ("Applied Formula: %s\n", expr);
   exprVarsParser[0]=10.0; exprVarsParser[1]=1.0;
   evalExprParser (expr); /* formula syntax test */
   exprVarsParser[0]=0.0; exprVarsParser[1]=0.0;
   //printf ("sizeof (rValuesTy):%u\n", valTy);
   //printf ("sizeof (rValues[%lu]):%lu\n", numV, val);
   //printf ("sizeof (resultsTy):%u\n", resTy);
   //printf ("sizeof (results[%lu]):%lu\n", totV, res);
   if (Eseries>0) {
      printf ("Resistors Series:E%u, Decades:%u\n", Eseries, decades);
      //printf ("Allocated about %lu MB of RAM\n", 220*powlmy(2, decades-5));
   } else {
      printf ("Resistors custom list of %u values\n", numR);
      /*maxRc = 2;
      maxRp = 1; // as now with custom list of values, we support only 1 resistor
      numR = listNumber;
      numV = listNumber;
      totV = numV * numV;*/
   }
   printf ("Allocated about %f MB of total RAM\n", allocatedMB);

   // 5 - fill the input vectors with needed data
   if (maxRp>1) {
      printf ("Calculating all:%u values with max:%u resistors per position ...\n", numV, maxRp);
   } else {
      printf ("Populating all:%u values with max:%u resistors per position ...\n", numV, maxRp);
   }
   ret = calcRvalues (); /* populate with all possible resistance values */
   if (ret==ERROR) {
      printf ("calcRvalues returned ERROR\n");
      free (results);
      return (ERROR);
   }

   // 6 - calculus of solutions
   printf ("Calculating all:%u solutions with a formula of:%u resistors ...\n", totV, maxRc);
   ret = calcFvalues (); /* calculate all results and delta */
   if (ret==ERROR) {
      printf ("calcFvalues returned ERROR\n");
      free (results);
      return (ERROR);
   }
   if (numBestRes>totV) numBestRes=totV;
   first = totV-numBestRes; /* 0 worse, 'first' 1st printed, 'totV' best, numBestRes # of best print */
//printf ("best:%u results are [%lu-%lu=best]\n", numBestRes, first, totV-1);
//printf ("Print last:%u founded results ...\n", numBestRes);
//ret = showVal (first);

   // 7 - sorting of solutions
   printf ("Sorting all:%u founded solutions ...\n", totV);
   structQuickSort (results, totV); /* QuickSort on all results */

   // 8 - print of results
   printf ("Printing best:%u solutions (top worst, botton best) in all configurations ...\n\n", numBestRes);
   if (maxRp==1) { // no need to showVal4,3,2 ...
      printf ("Show %u solutions with 2 resistors:\n", numBestRes);
      ret = showVal2 (numBestRes);
      printf ("\n");
      for (u32 e=0; e<numV; e++) {
         if (rValues[e].rp) free (rValues[e].rp);
      }
      if (results) free (results);
      return (OK);
   }
   printf ("Show %u solutions with up to 4 resistors:\n", numBestRes);
   ret = showVal (first);
   printf ("\n");
   printf ("Show %u solutions with 4 resistors:\n", numBestRes);
   ret = showVal4 (numBestRes);
   printf ("\n");
   printf ("Show %u solutions with 3 resistors:\n", numBestRes);
   ret = showVal3 (numBestRes);
   printf ("\n");
   printf ("Show %u solutions with 2 resistors:\n", numBestRes);
   ret = showVal2 (numBestRes);
   printf ("\n");

   // freeing dynamic allocated memory ...
   if (results) free (results);
   for (u32 e=0; e<numV; e++) {
      if (rValues[e].rp) free (rValues[e].rp);
   }
   if (rValues) free (rValues);
   if (baseR) free (baseR);
   if (expr) free (expr);
   return (OK);
}
